from typing import Any, Dict, Union

def validate_collection_metadata(metadata: Dict[Any, Any]) -> Dict[str, Union[str, int, float]]:
    if metadata is None:
        return metadata
    if len(metadata) == 0:
        return None
    # The new version of chroma only supports flat metadata. So we validate
    # that the metadata is is a flat dictionary of strings to ints, floats, or strings.
    for key, value in metadata.items():
        if not isinstance(value, (int, float, str)):
            raise ValueError(f"Metadata value {value} is not an int, float, or string. The new version of chroma only supports flat metadata. Please flatten your metadata using collection.modify() and try again.")
    return metadata

def migrate_embedding_metadata(metadata):
    return metadata if isinstance(metadata, dict) and len(metadata) > 0 else None



from typing import Dict
import clickhouse_connect
from chromadb.api import API
from chromadb.api.models.Collection import Collection
from tqdm import tqdm
import json

from chroma_migrate.utils import migrate_embedding_metadata, validate_collection_metadata


def migrate_from_clickhouse(api: API, host: str, port: int):
    conn = clickhouse_connect.get_client(
            host=host,
            port=port,
        )

    print("Loading existing collections...")
    # Read the collections from clickhouse
    collections = conn.query("SELECT uuid, name, metadata FROM collections").result_rows

    if len(collections) == 0:
        print("No collections found, exiting...")
        return False

    print("Validating collection metadata...")
    from_collection_to_metadata = {}
    for collection in collections:
        metadata = json.loads(collection[2])
        from_collection_to_metadata[collection[1]] = validate_collection_metadata(metadata)

    # Create the collections in chromadb
    print("Migrating existing collections...")
    collection_uuid_to_chroma_collection: Dict[str, Collection] = {}
    for collection in collections:
        uuid = collection[0]
        name = collection[1]
        metadata = from_collection_to_metadata[name]
        coll = api.get_or_create_collection(name, metadata)
        collection_uuid_to_chroma_collection[uuid] = coll

    # -------------------------------------
    
    # Add the embeddings to the collections
    print("Migrating existing embeddings...")
    with tqdm(total=conn.query("SELECT count(*) FROM embeddings").result_rows[0][0]) as pbar:
        with conn.query_row_block_stream('SELECT uuid, collection_uuid, id, embedding, document, metadata FROM embeddings') as stream:
            for block in stream:
                for record in block:
                    uuid = record[0]
                    collection_uuid = record[1]
                    id = record[2]
                    embedding = record[3]
                    document = record[4]
                    metadata = json.loads(record[5])
                    metadata = migrate_embedding_metadata(metadata)
                    collection = collection_uuid_to_chroma_collection[collection_uuid]
                    collection.add(id, embedding, metadata, document)
                    pbar.update(1)
    
    print(f"Migrated {len(collections)} collections and {pbar.n} embeddings")
    return True


from typing import Dict
from chromadb.api import API
from chromadb.api.models.Collection import Collection
from tqdm import tqdm
from more_itertools import chunked

from chroma_migrate.utils import migrate_embedding_metadata, validate_collection_metadata

CHUNK_SIZE = 1000

def migrate_from_remote_chroma(from_api: API, to_api: API):
    
    print("Loading Existing Collections...")
    from_collections = from_api.list_collections()

    if len(from_collections) == 0:
        print("No collections found, exiting...")
        return False

    print("Validating collection metadata...")
    from_collection_to_metadata = {}
    for collection in from_collections:
        from_collection_to_metadata[collection.name] = validate_collection_metadata(collection.metadata)

    print("Migrating existing collections...")
    from_collection_to_to_collection: Dict[str, Collection] = {}
    total_embeddings = 0
    for from_collection in from_collections:
        to_collection = to_api.get_or_create_collection(from_collection.name, from_collection_to_metadata[from_collection.name])
        total_embeddings += from_collection.count()
        from_collection_to_to_collection[from_collection.name] = to_collection
    
    print("Migrating existing embeddings...")
    with tqdm(total=total_embeddings) as pbar:
        for from_collection in from_collections:
            to_collection = from_collection_to_to_collection[from_collection.name]
            data = from_collection.get(include=["documents", "metadatas", "embeddings"])
            chunk_size = min(CHUNK_SIZE, len(data["ids"]))
            absolute_position = 0
            for chunk in chunked(data["ids"], chunk_size):
                ids = chunk
                embeddings = data["embeddings"][absolute_position:absolute_position+chunk_size]
                metadatas = data["metadatas"][absolute_position:absolute_position+chunk_size]
                for i, metadata in enumerate(metadatas):
                    metadatas[i] = migrate_embedding_metadata(metadata)

                documents = data["documents"][absolute_position:absolute_position+chunk_size]
                to_collection.add(ids, embeddings, metadatas, documents)
                pbar.update(len(chunk))
                absolute_position += len(chunk)
    
    print(f"Migrated {len(from_collections)} collections and {total_embeddings} embeddings")
    return True


# This example shows how to use chroma_migrate to migrate from any chroma database accessed via HTTP to a chroma database stored locally

from chroma_migrate.import_chromadb import migrate_from_remote_chroma

# Instantiate a chroma client that we wish to migrate to
import chromadb
api = chromadb.PersistentClient(path="./chroma_data")

# Migrate from a chroma server to the chroma client, assuming our old data was stored in the chroma server at localhost:8000
old_api = chromadb.HttpClient(host="localhost", port=8000)
migrate_from_remote_chroma(old_api, api)


import numpy as np
import random

import chromadb
from chromadb.api import API
from chromadb.config import Settings

# Run this to generate a test duckdb database
def gen():
    persist_directory_a = "./test_data_duckdb"
    
    # Create a new API
    api: API = chromadb.Client(Settings(chroma_db_impl="duckdb+parquet", persist_directory=persist_directory_a))

    # Create a random set of collections
    for i in range(10):
        collection = api.get_or_create_collection(f"collection_{i}")
        # Add a random set of embeddings
        N = random.randint(1, 100)
        D = random.randint(10, 100)
        embeddings = np.random.rand(N, D).tolist()
        metadata = [{f"{i}": f"{i}"} for i in range(N)]
        documents = [f"document_{i}" for i in range(N)]
        ids = [f"id_{i}" for i in range(N)]
        collection.add(ids, embeddings, metadata, documents)
        print(f"Added {N} embeddings to collection {collection.name}")
    
    api.persist()

if __name__ == "__main__":
    gen()


# This example shows how to use chroma_migrate to migrate from a chroma database stored in clickhouse to a chroma database stored locally

from chroma_migrate.import_clickhouse import migrate_from_clickhouse

# Instantiate a chroma client that we wish to migrate to
import chromadb
api = chromadb.PersistentClient(path="./chroma_data")

# Migrate from clickhouse to the chroma client, assuming our old data was stored in the clickhouse server at localhost:8123
migrate_from_clickhouse(api, "localhost", 8123)